/*
 * ProFTPD: mod_tds -- Support for connecting to MS SqlServer databases.
 * 
 * Patrick Muldoon <doon@inoc.net>
 * Based on mod_sql_mysql  by Johnnie Ingram 
 * 
 * THIS program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 */

#define MOD_SQL_TDS_VERSION "mod_sql_tds/0.5"

/* -- DO NOT MODIFY THE LINE BELOW UNLESS YOU FEEL LIKE IT --
 * $Libraries: -lct $
 */

/* This is mod_tds, contrib software for proftpd 1.2.0pre3 and above.
   For more information contact Patrick Muldoon <doon@inoc.net> 

   History Log:
x   
   * 2001-08-06: Convert to TDS Format 

*/

#include "conf.h"
#include <ctpublic.h>

/******************************************
 * Declarations of tds commands 
 ******************************************/
CS_RETCODE ct_con_drop(CS_CONNECTION *);
CS_RETCODE cs_ctx_drop(CS_CONTEXT *);
CS_RETCODE cs_ctx_alloc(CS_INT, CS_CONTEXT **);
CS_RETCODE ct_cmd_alloc(CS_CONNECTION *, CS_COMMAND **);
CS_RETCODE ct_cmd_drop(CS_COMMAND *);
CS_RETCODE ct_con_alloc(CS_CONTEXT *, CS_CONNECTION **);
CS_RETCODE ct_close(CS_CONNECTION *, CS_INT);
CS_RETCODE ct_init(CS_CONTEXT *, CS_INT);
CS_RETCODE ct_connect(CS_CONNECTION *, CS_CHAR*, CS_INT);
CS_RETCODE ct_con_props(CS_CONNECTION *, CS_INT,CS_INT,CS_VOID *,CS_INT,CS_INT *);
CS_RETCODE ct_command(CS_COMMAND *, CS_INT, CS_VOID *, CS_INT, CS_INT);
CS_RETCODE ct_results(CS_COMMAND *, CS_INT *);
CS_RETCODE ct_res_info(CS_COMMAND *, CS_INT, CS_VOID *, CS_INT, CS_INT);
CS_RETCODE ct_send(CS_COMMAND *);
CS_RETCODE ct_fetch(CS_COMMAND *, CS_INT, CS_INT, CS_INT,CS_INT *);
CS_RETCODE ct_bind(CS_COMMAND *, CS_INT, CS_DATAFMT *, CS_VOID *, CS_INT *, CS_SMALLINT *);

CS_CONTEXT    *context;       /* Context Structure */
CS_CONNECTION *connection;    /* Connection Structure */

struct sqldata_struc {
  int rowcount;
  int fieldcount;
  char **data;
};

struct columndata{
  CS_INT valuelen;
  char *value;
  CS_SMALLINT indicator;
};

struct tempdata{
  char **data;
  struct tempdata *next;
};

typedef struct tempdata tempdata_t;
typedef struct columndata columndata_t; 
typedef struct sqldata_struc sqldata_t;

/* Maximum username field to expect, etc. */
#define ARBITRARY_MAX                   128

static struct
{
  char *sql_host;   /* Reference to the Server name in Interface file  */
  char *sql_user;   /* What user we connect as                         */
  char *sql_pass;   /* What is our Password                            */
  char *sql_dbname; /* What database are we looking at                 */
  
  int ok;
  int opens;
} g;

/*
** Checks the return value of calls to the database
*/
int check_ret( CS_RETCODE ret, char* err_str ){
  if (ret != CS_SUCCEED){
    log_pri(LOG_ERR, "%s: %s on %s@%s",
	    MOD_SQL_TDS_VERSION, err_str);
    return FALSE;
  }else{
    return TRUE;
  }
  
}

/* **************************************************************** */

MODRET sql_cmd_close(cmd_rec * cmd)
{
  CS_RETCODE ret;
  log_debug(DEBUG5, "%s: close [%i] for %s", MOD_SQL_TDS_VERSION, g.opens,
            cmd->argv[0]);

  if (!g.ok || --g.opens)
    return DECLINED(cmd);
  
  
  if (!(context =  (CS_CONTEXT *) NULL)) {
    log_debug(DEBUG4, "%s: disconnecting", MOD_SQL_TDS_VERSION);
    /* close our connection */
    ret = ct_close(connection,CS_UNUSED);
    check_ret(ret, "Couldn't close connection");
    
    /* drop the control structure */
    ret = ct_con_drop(connection);
    check_ret(ret, "Couldn't Drop the Connection");
    
    /* drop the context */
    ret=cs_ctx_drop(context);
    check_ret(ret,"Couldn't Drop the Context");
  }
  /* null out the context */ 
  context = (CS_CONTEXT *) NULL;
  return HANDLED(cmd);
}

/*
** Open a connection to the database
*/
MODRET sql_cmd_open(cmd_rec * cmd)
{
  CS_RETCODE ret;

  if (!g.ok)
    return DECLINED(cmd);
  
  if (g.opens > 0)
    return HANDLED(cmd);

  g.opens++;
  
  log_debug(DEBUG5, "%s: open [%i] for %s", MOD_SQL_TDS_VERSION, g.opens,
            cmd->argv[0]);
  
  /* need to connect to the SQL server here  
  ** First initalize the App  
  ** Allocate a context Structure 
  */
  
  context = (CS_CONTEXT *) NULL;
  
  ret = cs_ctx_alloc(CS_VERSION_100, &context);
  
  if(!check_ret(ret, "Couldn't allocate context")){
    g.ok = FALSE;
    g.opens = 0;
    return DECLINED(cmd);
  }
  
  /* Initialize the Library */ 
  ret = ct_init(context, CS_VERSION_100);
  if (!check_ret(ret, "Couldn't Init Library")){
    g.ok = FALSE;
    g.opens = 0;
    return DECLINED(cmd);
  }
  
  /* allocate the connection structure */
  ret = ct_con_alloc(context, &connection);
  if(!check_ret(ret,"couldn't allocate a connection")){
    g.ok = FALSE;
    g.opens = 0;
    return DECLINED(cmd);
  }
  
  /* Set the Credentials */
  
  ret = ct_con_props(connection, CS_SET, CS_USERNAME, 
		     g.sql_user, CS_NULLTERM, NULL);
  if(!check_ret(ret, "Could not set user name")){
    g.ok = FALSE;
    g.opens= 0;
    return DECLINED(cmd);
  }
  ret = ct_con_props(connection, CS_SET, CS_PASSWORD, 
		     g.sql_pass, CS_NULLTERM, NULL);
  if(!check_ret(ret, "Could not set password")){
    g.ok = FALSE;
    g.opens = 0;
    return DECLINED(cmd);
  }
  
  /* Finally connect to the server  -- MySQL is sooo much easier*/
  ret = ct_connect(connection, g.sql_host, CS_NULLTERM);
  if(!check_ret(ret, "Couldn't connect!")){
    g.ok = FALSE;
    g.opens = 0;
    return DECLINED(cmd);
  }
  log_debug(DEBUG5, "%s: connect OK (%s@%s)", MOD_SQL_TDS_VERSION,
	    g.sql_user, g.sql_host);
  return HANDLED(cmd);
}

MODRET _do_query(cmd_rec * cmd, const char *query, int update){
  
  CS_RETCODE ret;
  CS_COMMAND *sqlcmd;
  
  log_debug(DEBUG5, "%s :  in _do_query , query=%s ", MOD_SQL_TDS_VERSION,query );
  
  if (!g.ok)
    return DECLINED(cmd);
  
  block_signals();
  
  /* allocate a command structure */
  ret =  ct_cmd_alloc(connection, &sqlcmd);
  
  /* init a lang command */
  ret = ct_command(sqlcmd, CS_LANG_CMD, &query, CS_NULLTERM, CS_UNUSED);
  log_debug(DEBUG5,"%s : _do_query : query = %s",MOD_SQL_TDS_VERSION, query);
  
  /* send the damn thing */
  ret = ct_send(sqlcmd);
  unblock_signals();

  if (ret != CS_SUCCEED) {
    log_debug(DEBUG4, "%s: %s failed: %s", MOD_SQL_TDS_VERSION,
              (update) ? "update" : "select", query);
    return DECLINED(cmd);
  }
  
  ret=ct_cmd_drop(sqlcmd);
  log_debug(DEBUG5, "%s: %s OK: [%s] %s", MOD_SQL_TDS_VERSION,
            (update) ? "update" : "select", g.sql_dbname, query);
  return HANDLED(cmd);
}

MODRET sql_cmd_update(cmd_rec * cmd)
{
  return _do_query(cmd, cmd->argv[1], TRUE);
}

MODRET sql_cmd_insert(cmd_rec * cmd)
{
  return _do_query(cmd, cmd->argv[1], TRUE);
}   

MODRET sql_cmd_select(cmd_rec * cmd)
{
  MODRET mr;
  int j,count;
  CS_RETCODE ret;              /*    Return Values                   */ 
  CS_RETCODE results_ret;      /*    Return Values                   */
  CS_INT result_type;          /*    Result type of Fetch            */
  CS_COMMAND *sqlcmd;          /*    cmd structure                   */ 
  CS_DATAFMT *datafmt;         /*    Data Format Structure           */
  columndata_t *coldata;       /*    What is actually in the columns */
  char **data;                 /*    The Data we send back           */                 
  CS_INT rcount;               /*    # of rows of data               */
  CS_INT fcount;               /*    # of fields per row             */
  CS_INT retcount;             /*    Count of rows returned          */
  sqldata_t *sd;               /*    Our sqldata Structure           */
  tempdata_t *td,*ptr;          /*    Holds are data till the end     */
  
  rcount = 0;
  count = 0;
  data = NULL;
  td = NULL; 
  
  /* We need to do the  following
   * 1. Build the Command Structure
   * 2. Send command to the server
   * 3. Determine # of columns in result 
   * 4. Loop through result, bind the fields, the fetch  into data
   */
  log_debug(DEBUG5,"In Sql_cmd_select: Query: %s",cmd->argv[1]);
  block_signals();
  
  /* allocate a command structure */
  ret =  ct_cmd_alloc(connection, &sqlcmd);
  if(!check_ret(ret, "Couldn't allocate Command Structure")){
    return DECLINED(cmd);
  }
  /* init a lang command */
  ret = ct_command(sqlcmd, CS_LANG_CMD, cmd->argv[1], CS_NULLTERM, CS_UNUSED);
  if(!check_ret(ret,"Couldn't init cmd")){
    return DECLINED(cmd);
  }
  
  /* send the damn thing */
  ret = ct_send(sqlcmd);
  unblock_signals(); 
  
  if(ret == CS_SUCCEED){
    
    while( (results_ret = ct_results(sqlcmd, &result_type)) == CS_SUCCEED){
      
      switch((int)result_type){
	
      case CS_ROW_RESULT:
	/*
	** Looks like this command returned rows
	*/
	
	ret = ct_res_info(sqlcmd, CS_NUMDATA, &fcount, CS_UNUSED, NULL);
	
	/* Allocate some space to hold the data we return */ 
	
	coldata = (columndata_t *)pcalloc(cmd->tmp_pool, 
					  (sizeof(columndata_t) * fcount));
	datafmt = (CS_DATAFMT *) pcalloc(cmd->tmp_pool, 
					 (sizeof(CS_DATAFMT) * fcount));
	
	/* TODO: ErrorCheck */
	
	/* Setup up the datafmt and prepare the coldata arrays */
	for (j = 0; j < fcount; j++){
	  
	  /* Set the max length to ARBITRARY_MAX */ 
	  datafmt[j].maxlength =  ARBITRARY_MAX;
	  
	  /* convert everything to null term strings */
	  datafmt[j].datatype = CS_CHAR_TYPE;
	  datafmt[j].format = CS_FMT_NULLTERM;
	  coldata[j].value = (char *)palloc(cmd->tmp_pool,
					    datafmt[j].maxlength);
	  /* Bind the columns */
	  ret = ct_bind(sqlcmd, (j+1), 
			&datafmt[j], 
			coldata[j].value, 
			&coldata[j].valuelen, 
			&coldata[j].indicator);
	}
	
	/* create a temp data structure*/ 
	td = (tempdata_t *)pcalloc(cmd->tmp_pool, sizeof(tempdata_t));
	/* make sure td points to the top of the list */
	
	ptr = td; 
	/* allocate some space to stor the data array */ 
	ptr->data = pcalloc(cmd->tmp_pool,(sizeof(char *) * fcount));
	ptr->next = NULL;
	
	while(((ret = ct_fetch(sqlcmd, CS_UNUSED, CS_UNUSED,
			       CS_UNUSED, &retcount))
	       == CS_SUCCEED)
	      || (ret == CS_ROW_FAIL)){
	  /*
	  ** Check if we hit a recoverable error.
	  */
	  if( ret == CS_ROW_FAIL ){
	    log_debug(DEBUG5,"%s Error on row %ld.",MOD_SQL_TDS_VERSION, retcount+1);
	  }
	  
	  
	  if(rcount > 0){
	    ptr->next = (tempdata_t *) pcalloc(cmd->tmp_pool, sizeof(tempdata_t));
	    ptr = ptr->next; 
	    ptr->data = pcalloc(cmd->tmp_pool,(sizeof(char *) * fcount));
	    ptr->next = NULL; 
	    log_debug(DEBUG5, "Created a  new temp record");
	  }
	  
	  /* add onto our temp record */ 
	  for(j=0;j<fcount;j++){
	    ptr->data[j] = pstrdup(cmd->tmp_pool, coldata[j].value);
	    log_debug(DEBUG5,"fetched field %d with value %s",j,ptr->data[j]);
	  }
	  rcount++; /* done with this row -- inc to the next */ 
	}
	
	if ( ret == CS_END_DATA ){
	  log_debug(DEBUG5,"All Done Processing Rows");
	}else{ 
	  /* Failure */
	  log_debug(DEBUG5, "error occured in fetch ret = %d",ret);
	  return DECLINED(cmd);
	}
	break;
	
      case CS_CMD_SUCCEED:
	/*
	** Command returned no rows
	*/
	log_debug(DEBUG5,"Command Returned no rows");
	break;
	
      case CS_CMD_FAIL:
	break;
      case CS_CMD_DONE:
	break;
      default:
	/*
	** We got something we weren't planning on
	*/
	log_debug(DEBUG5, "Unknown Result type %d returned",result_type);
	break;
	
      } /* End Switch */ 
    }  

    /* need to populate the data array */
    log_debug(DEBUG5,"Populate Data Array with %d rows of %d fields",rcount,fcount);
    data = pcalloc(cmd->tmp_pool, sizeof(char *) * ((rcount * fcount) + 1));
    
    ptr = td;
    count = 0; 
    while (ptr != NULL){
      for(j=0;j<fcount;j++){
	data[count++] = pstrdup(cmd->tmp_pool, ptr->data[j]);
	log_debug(DEBUG5,"copied ptr->data[%d] to data[%d] with value %s",j,count,ptr->data[j]);
      }
      ptr = ptr->next;
    }
        
    /* need to drop our cmd structure */
    ret = ct_cmd_drop(sqlcmd);
    log_debug(DEBUG5,"Dropped our command Structure");
    data[count] = NULL;
    log_debug(DEBUG5,"Set data[%d] to NULL",count);
    sd = (sqldata_t *) pcalloc(cmd->tmp_pool, sizeof(sqldata_t));
    log_debug(DEBUG5, "Allocated the sd struct");
    sd->rowcount = rcount;
    log_debug(DEBUG5,"set sd->rowcount to %d",rcount);
    sd->fieldcount = fcount;
    log_debug(DEBUG5,"set sd->fieldcount to %d",fcount);
    sd->data = data;
    log_debug(DEBUG5, "Set sd->data");
    mr = HANDLED(cmd); /* see if this works */
    mr->data = sd;
  }
  return mr;
}

static authtable tds_authtab[] = {
  {0, "dbd_open", sql_cmd_open},
  {0, "dbd_close", sql_cmd_close},
  {0, "dbd_update", sql_cmd_update},
  {0, "dbd_select", sql_cmd_select},
  {0, "dbd_insert", sql_cmd_insert},

  {0, NULL, NULL}
};

/* **************************************************************** */

static int tds_modinit()
{
  return 0;
}

static int tds_childinit()
{
  config_rec *c = NULL;
  char *info = NULL;
  char *pass = NULL;
  char *user = NULL;

  char *db = NULL;
  char *host = NULL;
  
  char *havehost = NULL;
  
  c = find_config(CURRENT_CONF, CONF_PARAM, "SQLConnectInfo", FALSE);

  if (!c) {
    /*
     * should log 
     */
    return 0;
  }

  info = c->argv[0];
  user = c->argv[1];
  pass = c->argv[2];

  db = pstrdup(session.pool, info);
  
  /* This is handled in the TDS Insterfaces File
   * we don't need the host:port combo
   * Look for the format db@server
   * Where server is in the Interfaces File
   */ 
  
  havehost = strchr(db, '@');
  
  if (havehost) {
    host = havehost + 1;
    *havehost = '\0';
  } else {
    host = "localhost";
  }
 
  
  memset(&g, 0, sizeof(g));

  g.sql_user = pstrdup(session.pool, user);
  g.sql_pass = pstrdup(session.pool, pass);

  g.sql_host = host;
  g.sql_dbname = db;
  //g.sql_port = atoi(port);

  g.ok = TRUE;

  log_debug(DEBUG5, "%s: configured: %s@%s for %s",
            MOD_SQL_TDS_VERSION, g.sql_dbname, g.sql_host,
            g.sql_user);
  return 0;
}

module sql_tds_module = {
  NULL, NULL,                   /* Always NULL */
  0x20,                         /* API Version 2.0 */
  "tds",                        /* We are called tds */ 
  NULL,                         /* SQL configuration handler table */
  NULL,                         /* SQL command handler table */
  tds_authtab,                /* SQL authentication handler table */
  tds_modinit,                /* Pre-fork "parent-mode" init */
  tds_childinit               /* Post-fork "child mode" init */
};

unsigned int sql_backend_escape_string(char *to, char *from, unsigned int len)
{
  strncpy(to, from, len);
  to[len] = '\0';
  return strlen(from);
}

int sql_backend_check_auth(cmd_rec * cmd, const char *c_clear,
                           const char *c_hash){
  log_debug(DEBUG5, "%s : WTF? We don't do this sql_backend_auth : clear %s, hash %s",MOD_SQL_TDS_VERSION,c_clear,c_hash);
  return 0; 
}
